/* Copyright (C) 1997-1999, Red Hat Software
 * This may be freely redistributed under the terms of the GNU Public License.
 *
 * Tiny portions copyright 2001-2002, BSI
 */

#include <config.h>

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <glob.h>
#include <libintl.h> 
#include <locale.h> 

#if (!HAVE_LIBPOPT && !HAVE_GETOPT_LONG && !HAVE_LIBGETOPT)
    #error "Either the Redhat popt or GNU getopt library must be installed."
#endif

#if HAVE_LIBPOPT
    #include <popt.h>
#else /* !HAVE_LIBPOPT */
    /* we need GNU getopt-long */
    #define _GNU_SOURCE
    #include <getopt.h>
#endif /* HAVE_LIBPOPT */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
static char *progname;
extern int addItem;
struct config conf;		// used by leveldb.c

#define _(String) gettext((String)) 

#include "leveldb.h"

static void usage(void) {
    fprintf(stderr, _("%s version %s - Copyright (C) 1997-2000 Red Hat, Inc.\n"), progname, VERSION);
    fprintf(stderr, _("This may be freely redistributed under the terms of "
			"the GNU Public License.\n"));
    fprintf(stderr, "\n");
    fprintf(stderr, _("usage:   %s --list [name]\n"), progname);
    fprintf(stderr, _("         %s --add <name>\n"), progname);
    fprintf(stderr, _("         %s --del <name>\n"), progname);
    fprintf(stderr, _("         %s [--level <levels>] <name> %s\n"), progname, "<on|off|reset>");

    exit(1);
}

static void readServiceError(int rc, char * name) {
    if (rc == 1) {
	fprintf(stderr, _("service %s does not support chkconfig\n"), name);
    } else {
	fprintf(stderr, _("error reading information on service %s: %s\n"),
		name, strerror(errno));
    }

    exit(1);
}

static int delService(char * name) {
    int level, i, rc;
    glob_t globres;
    struct service s;

    if ((rc = readServiceInfo(name, &s, 0))) {
	readServiceError(rc, name);
	return 1;
    }

    for (level = 0; level <= conf.maxlevel; level++) {
	    if (!findServiceEntries(name, level, &globres)) {
		    for (i = 0; i < globres.gl_pathc; i++)
		      unlink(globres.gl_pathv[i]);
		    if (globres.gl_pathc) globfree(&globres);
	    }
    }
    return 0;
}

static int addService(char * name) {
    int i, rc;
    struct service s;

    if ((rc = readServiceInfo(name, &s, 0))) {
	readServiceError(rc, name);
	return 1;
    }
    
    for (i = 0; i <= conf.maxlevel; i++) {
	    if ((1 << i) & s.levels)
		doSetService(s, i, 1);
	    else
		doSetService(s, i, 0);
    }

    return 0;
}

static int showServiceInfo( char * name, int forgiving)
{
    int rc;
    int i;
    struct service s;
	char *on, *off;

    if ((rc = readServiceInfo(name, &s, 0))) {
	if (!forgiving)
	    readServiceError(rc, name);
	return forgiving ? 0 : 1;
    }

    printf("%-15s", s.name);

	on = getenv("CHKCONFIG_ON");
	off = getenv("CHKCONFIG_OFF");

    for (i = 0; i <= conf.maxlevel; i++)
		printf("\t%d:%s", i, isOn(s.name, i) ? on ? on : _("on") : off ? off : _("off"));
    
	printf("\n");

    return 0;
}

static int listService(char * item) {
    DIR * dir;
    struct dirent * ent;
    struct stat sb;
    char fn[1024];
    int err = 0;

    if (item) return showServiceInfo(item, 0);

    if (!(dir = opendir(conf.initdir))) {
	fprintf(stderr, _("failed to open %s: %s\n"), conf.initdir,
		strerror(errno));
        return 1;
    }

    while ((ent = readdir(dir))) {
	const char *dn;

	/* Skip any file starting with a . */
	if (ent->d_name[0] == '.')	continue;

	/* Skip files with known bad extensions */
	if ((dn = strrchr(ent->d_name, '.')) != NULL &&
    (!strcmp(dn, ".rpmsave") || !strcmp(dn, ".rpmnew")
     || !strcmp(dn, ".rpmorig") || !strcmp(dn, ".swp")
     || !strcmp(dn, ".dpkg-dist") || !strcmp(dn, ".dpkg-old")
     || !strcmp(dn, ".dpkg-new")))
	    continue;

	dn = ent->d_name + strlen(ent->d_name) - 1;
	if (*dn == '~' || *dn == ',')
	    continue;
	
	sprintf(fn, "%s/%s", conf.initdir, ent->d_name);
	if (stat(fn, &sb)) {
	    err = errno;
	    continue;
	}
	if (!S_ISREG(sb.st_mode)) continue;

	if (showServiceInfo(ent->d_name, 1)) {
	    closedir(dir);
	    return 1;
	}
    }

    if (err) {
	fprintf(stderr, _("error reading from directory %s: %s"), 
		conf.initdir, strerror(err));
        return 1;
    }

    closedir(dir);

    return 0;
}

int setService(char * name, int where, int state) {
    int i, rc;
    int what;
    struct service s;
    
    if (!where && state != -1) {
	/* levels 3, 4, 5 */
	where = (1 << 3) | (1 << 4) | (1 << 5);
    } else if (!where) {
	where = (1 << 0) | (1 << 1) | (1 << 2) |
	        (1 << 3) | (1 << 4) | (1 << 5) | (1 << 6);
    }

    if ((rc = readServiceInfo(name, &s, 0))) {
	readServiceError(rc, name);
	return 1;
    }

	    for (i = 0; i <= conf.maxlevel; i++) {
		    if (!((1 << i) & where)) continue;

		    if (state == 1 || state == 0)
		      what = state;
		    else if (s.levels & (1 << i))
		      what = 1;
		    else
		      what = 0;
		    doSetService(s, i, what);
	    }

    return 0;
}


int main(int argc, char ** argv) {
    int listItem = 0, delItem = 0;
    int rc, i, x;
    char * levels = NULL;
    int help=0, version=0;

#if HAVE_LIBPOPT
    poptContext optCon;
    struct poptOption optionsTable[] = {
	    { "add", '\0', 0, &addItem, 0 },
	    { "del", '\0', 0, &delItem, 0 },
	    { "list", '\0', 0, &listItem, 0 },
	    { "level", '\0', POPT_ARG_STRING, &levels, 0 },
	    { "rc.d", '\0', POPT_ARG_STRING, &conf.rcdir, 0 },
	    { "init.d", '\0', POPT_ARG_STRING, &conf.initdir, 0 },
	    { "xinetd.d", '\0', POPT_ARG_STRING, &conf.xinetddir, 0 },
	    { "help", 'h', POPT_ARG_NONE, &help, 0 },
	    { "version", 'v', POPT_ARG_NONE, &version, 0 },
	    { 0, 0, 0, 0, 0 } 
    };
#else /* !HAVE_LIBPOPT */
    struct option optionsTable[] = 
    {
        { "add", 0, &addItem, 1 },
        { "del", 0, &delItem, 1 },
        { "list", 0, &listItem, 1 },
        { "level", 1, 0, 0 },
        { "rc.d", 1, 0, 0 },
        { "init.d", 1, 0, 0 },
        { "xinetd.d", 1, 0, 0 },
        { "help", 0, &help, 1 },
        { "version", 0, &version, 1 },
        { 0, 0, 0, 0 }
    };
#endif /* HAVE_LIBPOPT */

    conf.rcdir = RCDIR;		// default to value found by configure
    conf.initdir = INITDIR;
    conf.xinetddir = XINETDDIR;
    conf.maxlevel = MAXLEVEL;

    if ((progname = strrchr(argv[0], '/')) != NULL)
	progname++;
    else
	progname = argv[0];

    setlocale(LC_ALL, ""); 
    bindtextdomain(PACKAGE, LOCALEDIR); 
    textdomain(PACKAGE); 
#if HAVE_LIBPOPT
    optCon = poptGetContext("chkconfig", argc, argv, optionsTable, 0);
    poptReadDefaultConfig(optCon, 1);

    if ((rc = poptGetNextOpt(optCon)) < -1) {
	fprintf(stderr, "%s: %s\n", 
		poptBadOption(optCon, POPT_BADOPTION_NOALIAS), 
		poptStrerror(rc));
	exit(1);
    }
#else /* !HAVE_LIBPOPT */
    /* turn off getopt_long from reporting errors on stderr */
    opterr = 0;

    /* process all the options so that we know what the user wants */
    while ((rc = getopt_long(argc, argv, "hv",
				    optionsTable, &i)) != -1) {
	switch (rc) {
	    case 'h':	help = 1;	break;
	    case 'v':	version =1;	break;
	    case ':':	/* missing arg for option  -- fall through */
	    case '?':	/* extra arg for option */
			usage();	/* exits */
	    case 0:	/* found a long option... check for an arg */
			switch (i) {
			    /* these numbers correspond to the
			     * index in optionsTable[]. */
			    case 3: levels = optarg; break;
			    case 4: conf.rcdir = optarg; break;
			    case 5: conf.initdir = optarg; break;
			    case 6: conf.xinetddir = optarg; break;
			    default: break;	/* no arg */
			}
			break;
	    default:
		fprintf(stderr, "%s: weird, getopt_long = %d (%c), i = %d\n",
			progname, rc, (char) rc, i);
		fprintf(stderr, "Terminating.\n");
		exit(1);
	}
    }
#endif /* HAVE_LIBPOPT */

    if (version) {
	fprintf(stdout, _("%s version %s\n"), progname, VERSION);
	exit(0);
    }

    if (help || argc == 1) usage();

    if ((listItem + addItem + delItem) > 1) {
	fprintf(stderr, _("only one of --list, --add, or --del may be "
		"specified\n"));
	exit(1);
    }

#if !HAVE_LIBPOPT
    /* for getopt, we just use argv[optind].  This
     * trickiness preserves the original code, so that
     * we can do merges later. */
    #define poptGetArg(x)	((optind < argc) ? argv[optind++] : NULL)
#endif

    if (addItem) {
	char * name = (char *)poptGetArg(optCon);

	if (!name || !*name || poptGetArg(optCon))
	    usage();

	return addService(name);
    } else if (delItem) {
	char * name = (char *)poptGetArg(optCon);

	if (!name || !*name || poptGetArg(optCon)) usage();

	return delService(name);
    } else if (listItem) {
	char * item = (char *)poptGetArg(optCon);

	if (item && poptGetArg(optCon)) usage();

	return listService(item);
    } else {
	char * name = (char *)poptGetArg(optCon);
	char * state = (char *)poptGetArg(optCon);
	int where = 0, level = -1;

	if (levels) {
	    where = parseLevels(levels, 0);
	    if (where == -1) usage();
	}

	if (!state) {
	    if (where) {
		rc = x = 0;
		i = where;
		while (i) {
		    if (i & 1) {
			rc++;
			level = x;
		    }
		    i >>= 1;
		    x++;
		}

		if (rc > 1) {
		    fprintf(stderr, _("only one runlevel may be specified for "
			    "a chkconfig query\n"));
		    exit(1);
		}
	    } 

	    return isOn(name, level) ? 0 : 1;
	} else if (!strcmp(state, "on"))
	    return setService(name, where, 1);
	else if (!strcmp(state, "off"))
	    return setService(name, where, 0);
	else if (!strcmp(state, "reset"))
	    return setService(name, where, -1);
	else
	    usage();
    }

    usage();

    return 1;
}
